from dataclasses import dataclass
from typing import List, Dict, Optional, Literal


class SgRecord:
    """
    A canonical representation of a SafeGraph record
    """

    MISSING = "<MISSING>"

    @dataclass(frozen=True)
    class Headers:
        PAGE_URL            = "page_url"
        LOCATION_NAME       = "location_name"
        STREET_ADDRESS      = "street_address"
        CITY                = "city"
        STATE               = "state"
        ZIP                 = "zip"
        COUNTRY_CODE        = "country_code"
        STORE_NUMBER        = "store_number"
        PHONE               = "phone"
        LOCATION_TYPE       = "location_type"
        LATITUDE            = "latitude"
        LONGITUDE           = "longitude"
        LOCATOR_DOMAIN      = "locator_domain"
        HOURS_OF_OPERATION  = "hours_of_operation"
        RAW_ADDRESS         = "raw_address"
        RECORD_ID           = "record_id"

        HEADER_ROW = [PAGE_URL,
                      LOCATION_NAME,
                      STREET_ADDRESS,
                      CITY,
                      STATE,
                      ZIP,
                      COUNTRY_CODE,
                      STORE_NUMBER,
                      PHONE,
                      LOCATION_TYPE,
                      LATITUDE,
                      LONGITUDE,
                      LOCATOR_DOMAIN,
                      HOURS_OF_OPERATION,
                      RAW_ADDRESS]

        HEADER_ROW_WITH_REC_ID = HEADER_ROW + [RECORD_ID]

        HeaderUnion = Literal[PAGE_URL,         # noqa: these will never change
                              LOCATION_NAME,
                              STREET_ADDRESS,
                              CITY,
                              STATE,
                              ZIP,
                              COUNTRY_CODE,
                              STORE_NUMBER,
                              PHONE,
                              LOCATION_TYPE,
                              LATITUDE,
                              LONGITUDE,
                              LOCATOR_DOMAIN,
                              HOURS_OF_OPERATION,
                              RAW_ADDRESS]

    def __init__(self,
                 raw: Optional[Dict[str, str]] = None,
                 page_url: str = MISSING,
                 location_name: str = MISSING,
                 street_address: str = MISSING,
                 city: str = MISSING,
                 state: str = MISSING,
                 zip_postal: str = MISSING,
                 country_code: str = MISSING,
                 store_number: str = MISSING,
                 phone: str = MISSING,
                 location_type: str = MISSING,
                 latitude: str = MISSING,
                 longitude: str = MISSING,
                 locator_domain: str = MISSING,
                 hours_of_operation: str = MISSING,
                 raw_address: str = MISSING):
        """
        Constructs an SgRecord either from a dictionary (`raw`), or else from the by-name parameters.

        if `raw` is not `None`, fields will be populated using `raw`;
        otherwise, they will be populated using the by-name params.

        The dict keys in `raw` are keyed on `SgRecord.Headers`
        """

        if raw is not None:
            self.__from_dict(raw)
        else:
            self.__page_url = SgRecord.__or_missing(page_url)
            self.__location_name = SgRecord.__or_missing(location_name)
            self.__street_address = SgRecord.__or_missing(street_address)
            self.__city = SgRecord.__or_missing(city)
            self.__state = SgRecord.__or_missing(state)
            self.__zip_postal = SgRecord.__or_missing(zip_postal)
            self.__country_code = SgRecord.__or_missing(country_code)
            self.__store_number = SgRecord.__or_missing(store_number)
            self.__phone = SgRecord.__or_missing(phone)
            self.__location_type = SgRecord.__or_missing(location_type)
            self.__latitude = SgRecord.__or_missing(latitude)
            self.__longitude = SgRecord.__or_missing(longitude)
            self.__locator_domain = SgRecord.__or_missing(locator_domain)
            self.__hours_of_operation = SgRecord.__or_missing(hours_of_operation)
            self.__raw_address = SgRecord.__or_missing(raw_address)

        self.__as_row = self.__to_row()
        self.__as_dict = self.__to_dict()

    def __from_dict(self, raw: Dict[str, str]):
        """
        Constructs an SgRecord from a str->str dictionary that has the canonical names as keys.
        """
        self.__page_url = SgRecord.__or_missing(raw.get(SgRecord.Headers.PAGE_URL))
        self.__location_name = SgRecord.__or_missing(raw.get(SgRecord.Headers.LOCATION_NAME))
        self.__street_address = SgRecord.__or_missing(raw.get(SgRecord.Headers.STREET_ADDRESS))
        self.__city = SgRecord.__or_missing(raw.get(SgRecord.Headers.CITY))
        self.__state = SgRecord.__or_missing(raw.get(SgRecord.Headers.STATE))
        self.__zip_postal = SgRecord.__or_missing(raw.get(SgRecord.Headers.ZIP))
        self.__country_code = SgRecord.__or_missing(raw.get(SgRecord.Headers.COUNTRY_CODE))
        self.__store_number = SgRecord.__or_missing(raw.get(SgRecord.Headers.STORE_NUMBER))
        self.__phone = SgRecord.__or_missing(raw.get(SgRecord.Headers.PHONE))
        self.__location_type = SgRecord.__or_missing(raw.get(SgRecord.Headers.LOCATION_TYPE))
        self.__latitude = SgRecord.__or_missing(raw.get(SgRecord.Headers.LATITUDE))
        self.__longitude = SgRecord.__or_missing(raw.get(SgRecord.Headers.LONGITUDE))
        self.__locator_domain = SgRecord.__or_missing(raw.get(SgRecord.Headers.LOCATOR_DOMAIN))
        self.__hours_of_operation = SgRecord.__or_missing(raw.get(SgRecord.Headers.HOURS_OF_OPERATION))
        self.__raw_address = SgRecord.__or_missing(raw.get(SgRecord.Headers.RAW_ADDRESS))

    def page_url(self) -> str:
        return self.__page_url

    def location_name(self) -> str:
        return self.__location_name

    def street_address(self) -> str:
        return self.__street_address

    def city(self) -> str:
        return self.__city

    def state(self) -> str:
        return self.__state

    def zip_postal(self) -> str:
        return self.__zip_postal

    def country_code(self) -> str:
        return self.__country_code

    def store_number(self) -> str:
        return self.__store_number

    def phone(self) -> str:
        return self.__phone

    def location_type(self) -> str:
        return self.__location_type

    def latitude(self) -> str:
        return self.__latitude

    def longitude(self) -> str:
        return self.__longitude

    def locator_domain(self) -> str:
        return self.__locator_domain

    def hours_of_operation(self) -> str:
        return self.__hours_of_operation

    def raw_address(self) -> str:
        return self.__raw_address

    def as_row(self) -> List[str]:
        return self.__as_row

    def as_dict(self) -> Dict[str, Optional[str]]:
        return self.__as_dict

    def __to_dict(self) -> Dict[str, Optional[str]]:
        return dict(zip(SgRecord.Headers.HEADER_ROW, self.as_row()))

    def __to_row(self) -> List[str]:
        return [self.__page_url,
                self.__location_name,
                self.__street_address,
                self.__city,
                self.__state,
                self.__zip_postal,
                self.__country_code,
                self.__store_number,
                self.__phone,
                self.__location_type,
                self.__latitude,
                self.__longitude,
                self.__locator_domain,
                self.__hours_of_operation,
                self.__raw_address]

    def __str__(self) -> str:
        return str(self.__to_dict())

    @staticmethod
    def __or_missing(x: Optional[str]) -> str:
        return x if x else SgRecord.MISSING
