from typing import Set, Tuple

from .pause_resume import CrawlState
from .sgrecord import SgRecord
from .sgrecord_id import SgRecordID


class SgRecordDeduper:
    def __init__(self,
                 record_id: SgRecordID,
                 resume_if_paused: bool = True,
                 data_file_path: str = CrawlState.DEFAULT_DATA_FILE):
        """
        Deduplicates records by holding an internal set of record ids.

        :param record_id: The way by which to deduplicate a record.
        :param resume_if_paused: Reads the existing file, if present, and adds ids to the set of ids. [default=True]
        :param data_file_path: The path to the existing data file.
        """
        self.__id = record_id
        self.__encountered_ids: Set[str] = self.__read_existing_file(record_id, data_file_path) if resume_if_paused else set()

    def get_id(self) -> SgRecordID:
        """
        Returns the record ID class that's used
        """
        return self.__id

    @staticmethod
    def __read_existing_file(record_id: SgRecordID, data_file_path: str) -> Set[str]:
        ids = set()
        for rec in CrawlState.load_data(data_file_path):
            ids.add(record_id.generate_id(rec))
        return ids

    def dedup(self, record: SgRecord) -> Tuple[bool, str]:
        """
        Returns a tuple signifying: (is_duplicate?, generated_identity)
        """
        rec_id = self.__id.generate_id(record)
        if rec_id in self.__encountered_ids:
            return True, rec_id
        else:
            self.__encountered_ids.add(rec_id)
            return False, rec_id