import csv
import threading
from os import path
from typing import Optional

from sglogging import SgLogSetup

from sgscrape.pause_resume import CrawlState
from sgscrape.sgrecord import SgRecord
from sgscrape.sgrecord_id import RecommendedRecordIds
from sgscrape.sgrecord_deduper import SgRecordDeduper


class SgWriter:
    """
    Write records to a csv file safely and efficiently.
    When provided with an `SgRecordDeduper`, it also deduplicates
    based on current, and past (i.e. resumed) records.

    Usage:
    ```
    def fetch_data(sgw: SgWriter):
        sgw.write_row(SgRecord(page_url="123.com/1", store_number=''))
        sgw.write_row(SgRecord(raw={'page_url':"123.com/2", 'store_number': ''}))
        sgw.write_row(SgRecord(raw={'page_url':"123.com/3", 'store_number': ''}))

    with SgWriter(deduper=SgRecordDeduper(RecommendedRecordIds.PageUrlId)) as writer:
        fetch_data(writer)
    ```
    """

    __lock = threading.Lock()

    def __init__(self,
                 deduper: Optional[SgRecordDeduper] = None,
                 data_file: str = CrawlState.DEFAULT_DATA_FILE):
        """
        Creates the writer. Note that the constructor also writes the header row (if necessary).

        :param deduper: Optionally, inject a record deduplicator to filter out duplicates. If present, the data file will
                        contain an additional colon-delimited row that represents the record's unique id creation.
        """
        self.__log = SgLogSetup().get_logger(logger_name='sgwriter')
        self.__deduper = deduper
        self.__data_file_name = data_file
        preexisting = path.exists(data_file)
        write_mode = 'a' # if preexisting else 'w'
        self.__data_file = open(self.__data_file_name, mode=write_mode, encoding='utf-8')
        self.__writer = csv.writer(self.__data_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_ALL, lineterminator='\n')
        self.__id_str = str(deduper.get_id()) if deduper else None
        header_row = SgRecord.Headers.HEADER_ROW_WITH_REC_ID if self.__id_str else SgRecord.Headers.HEADER_ROW
        if not preexisting:
            self.__writer.writerow(header_row)
        else:
            self.__log.debug(f'Pre-existing data file found: {data_file} . Skipping header row.')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__data_file.close()

    def write_row(self, record: SgRecord) -> Optional[str]:
        """
        Writes the record to file, or else returns the id of a record, only if it is a duplicate.
        """
        if not self.__deduper:
            SgWriter.__lock.acquire()
            self.__writer.writerow(record.as_row())
            SgWriter.__lock.release()
            return None
        else:
            is_dup, id = self.__deduper.dedup(record)
            if not is_dup:
                row = record.as_row()
                row.append(self.__id_str)
                SgWriter.__lock.acquire()
                self.__writer.writerow(row)
                SgWriter.__lock.release()
                return None
            else:
                return id


if __name__ == "__main__":
    def fetch_data(sgw: SgWriter):
        sgw.write_row(SgRecord(page_url="123.com/1", store_number='', latitude=3.1415, longitude=3.1415))
        sgw.write_row(SgRecord(raw={'page_url':"123.com/2", 'store_number': ''}))
        sgw.write_row(SgRecord(raw={'page_url':"123.com/3", 'store_number': ''}))

    with SgWriter(deduper=SgRecordDeduper(RecommendedRecordIds.PageUrlId)) as writer:
        fetch_data(writer)

    # writes page_url as a string, and everything else, including `store_data`, as <MISSING>.